import gguf
import numpy as np
import torch
import torch.nn as nn
import torchvision.datasets as dsets
import torchvision.transforms as transforms
from torch.autograd import Variable

import sys
from time import time

input_size  = 784  # img_size = (28,28) ---> 28*28=784 in total
hidden_size = 500  # number of nodes at hidden layer
num_classes = 10   # number of output classes discrete range [0,9]
num_epochs  = 30   # number of times which the entire dataset is passed throughout the model
batch_size  = 1000 # the size of input data used for one iteration
lr          = 1e-3 # size of step


class Net(nn.Module):
    def __init__(self, input_size, hidden_size, num_classes):
        super(Net, self).__init__()
        self.fc1 = nn.Linear(input_size, hidden_size)
        self.relu = nn.ReLU()
        self.fc2 = nn.Linear(hidden_size, num_classes)

    def forward(self, x):
        out = self.fc1(x)
        out = self.relu(out)
        out = self.fc2(out)
        return out


def train(model_path):
    train_data = dsets.MNIST(root='./data', train=True, transform=transforms.ToTensor(), download=True)
    test_data  = dsets.MNIST(root='./data', train=False, transform=transforms.ToTensor())

    assert len(train_data) == 60000
    assert len(test_data)  == 10000

    kwargs_train_test = dict(batch_size=batch_size, num_workers=4, pin_memory=True)
    train_gen = torch.utils.data.DataLoader(dataset=train_data, shuffle=True,  **kwargs_train_test)
    test_gen  = torch.utils.data.DataLoader(dataset=test_data,  shuffle=False, **kwargs_train_test)

    net = Net(input_size, hidden_size, num_classes)

    if torch.cuda.is_available():
        net.cuda()

    loss_function = nn.CrossEntropyLoss()
    optimizer = torch.optim.Adam(net.parameters(), lr=lr)

    t_start = time()
    for epoch in range(num_epochs):
        loss_history = []
        ncorrect = 0

        for i, (images, labels) in enumerate(train_gen):
            images = Variable(images.view(-1, 28*28))
            labels = Variable(labels)

            if torch.cuda.is_available():
                images = images.cuda()
                labels = labels.cuda()

            optimizer.zero_grad()
            outputs = net(images)
            loss = loss_function(outputs, labels)

            loss_history.append(loss.cpu().data)
            _, predictions = torch.max(outputs, 1)
            ncorrect += (predictions == labels).sum()

            loss.backward()
            optimizer.step()

            if (i + 1)*batch_size % 10000 == 0:
                loss_mean = np.mean(loss_history)
                accuracy = ncorrect / ((i + 1) * batch_size)
                print(
                    f"Epoch [{epoch+1:02d}/{num_epochs}], "
                    f"Step [{(i+1)*batch_size:05d}/{len(train_data)}], "
                    f"Loss: {loss_mean:.4f}, Accuracy: {100*accuracy:.2f}%")
    print()
    print(f"Training took {time()-t_start:.2f}s")

    loss_history = []
    ncorrect = 0

    for i, (images, labels) in enumerate(test_gen):
        images = Variable(images.view(-1, 28*28))
        labels = Variable(labels)

        if torch.cuda.is_available():
            images = images.cuda()
            labels = labels.cuda()

        outputs = net(images)
        loss = loss_function(outputs, labels)

        loss_history.append(loss.cpu().data)
        _, predictions = torch.max(outputs, 1)
        ncorrect += (predictions == labels).sum().cpu().numpy()

    loss_mean            = np.mean(loss_history)
    loss_uncertainty     = np.std(loss_history) / np.sqrt(len(loss_history) - 1)
    accuracy_mean        = ncorrect / (len(test_gen) * batch_size)
    accuracy_uncertainty = np.sqrt(accuracy_mean * (1.0 - accuracy_mean) / (len(test_gen) * batch_size))
    print()
    print(f"Test loss: {loss_mean:.6f}+-{loss_uncertainty:.6f}, Test accuracy: {100*accuracy_mean:.2f}+-{100*accuracy_uncertainty:.2f}%")

    gguf_writer = gguf.GGUFWriter(model_path, "mnist-fc")

    print()
    print(f"Model tensors saved to {model_path}:")
    for tensor_name in net.state_dict().keys():
        data = net.state_dict()[tensor_name].squeeze().cpu().numpy()
        print(tensor_name, "\t", data.shape)
        gguf_writer.add_tensor(tensor_name, data)

    gguf_writer.write_header_to_file()
    gguf_writer.write_kv_data_to_file()
    gguf_writer.write_tensors_to_file()
    gguf_writer.close()


if __name__ == '__main__':
    if len(sys.argv) != 2:
        print(f"Usage: {sys.argv[0]} <model_path>")
        sys.exit(1)
    train(sys.argv[1])



#!/usr/bin/env python3
import sys
import gguf
import numpy as np

def save_conv2d_layer(f, gguf_writer, prefix, inp_c, filters, size, batch_normalize=True):
    biases = np.fromfile(f, dtype=np.float32, count=filters)
    gguf_writer.add_tensor(prefix + "_biases", biases, raw_shape=(1, filters, 1, 1))

    if batch_normalize:
        scales = np.fromfile(f, dtype=np.float32, count=filters)
        gguf_writer.add_tensor(prefix + "_scales", scales, raw_shape=(1, filters, 1, 1))
        rolling_mean = np.fromfile(f, dtype=np.float32, count=filters)
        gguf_writer.add_tensor(prefix + "_rolling_mean", rolling_mean, raw_shape=(1, filters, 1, 1))
        rolling_variance = np.fromfile(f, dtype=np.float32, count=filters)
        gguf_writer.add_tensor(prefix + "_rolling_variance", rolling_variance, raw_shape=(1, filters, 1, 1))

    weights_count = filters * inp_c * size * size
    l0_weights = np.fromfile(f, dtype=np.float32, count=weights_count)
    ## ggml doesn't support f32 convolution yet, use f16 instead
    l0_weights = l0_weights.astype(np.float16)
    gguf_writer.add_tensor(prefix + "_weights", l0_weights, raw_shape=(filters, inp_c, size, size))


if __name__ == '__main__':
    if len(sys.argv) != 2:
        print("Usage: %s <yolov3-tiny.weights>" % sys.argv[0])
        sys.exit(1)
    outfile = 'yolov3-tiny.gguf'
    gguf_writer = gguf.GGUFWriter(outfile, 'yolov3-tiny')

    f = open(sys.argv[1], 'rb')
    f.read(20) # skip header
    save_conv2d_layer(f, gguf_writer, "l0", 3, 16, 3)
    save_conv2d_layer(f, gguf_writer, "l1", 16, 32, 3)
    save_conv2d_layer(f, gguf_writer, "l2", 32, 64, 3)
    save_conv2d_layer(f, gguf_writer, "l3", 64, 128, 3)
    save_conv2d_layer(f, gguf_writer, "l4", 128, 256, 3)
    save_conv2d_layer(f, gguf_writer, "l5", 256, 512, 3)
    save_conv2d_layer(f, gguf_writer, "l6", 512, 1024, 3)
    save_conv2d_layer(f, gguf_writer, "l7", 1024, 256, 1)
    save_conv2d_layer(f, gguf_writer, "l8", 256, 512, 3)
    save_conv2d_layer(f, gguf_writer, "l9", 512, 255, 1, batch_normalize=False)
    save_conv2d_layer(f, gguf_writer, "l10", 256, 128, 1)
    save_conv2d_layer(f, gguf_writer, "l11", 384, 256, 3)
    save_conv2d_layer(f, gguf_writer, "l12", 256, 255, 1, batch_normalize=False)
    f.close()
    
    gguf_writer.write_header_to_file()
    gguf_writer.write_kv_data_to_file()
    gguf_writer.write_tensors_to_file()
    gguf_writer.close()
    print("{} converted to {}".format(sys.argv[1], outfile))



import pytest
from pytest import raises

from ggml import lib, ffi
from ggml.utils import init, copy, numpy
import numpy as np
import numpy.testing as npt

@pytest.fixture()
def ctx():
    print("setup")
    yield init(mem_size=10*1024*1024)
    print("teardown")

class TestNumPy:
    
    # Single element

    def test_set_get_single_i32(self, ctx):
        i = lib.ggml_new_i32(ctx, 42)
        assert lib.ggml_get_i32_1d(i, 0) == 42
        assert numpy(i) == np.array([42], dtype=np.int32)

    def test_set_get_single_f32(self, ctx):
        i = lib.ggml_new_f32(ctx, 4.2)
        
        epsilon = 0.000001 # Not sure why so large a difference??
        pytest.approx(lib.ggml_get_f32_1d(i, 0), 4.2, epsilon)
        pytest.approx(numpy(i), np.array([4.2], dtype=np.float32), epsilon)

    def _test_copy_np_to_ggml(self, a: np.ndarray, t: ffi.CData):
        a2 = a.copy() # Clone original
        copy(a, t)
        npt.assert_array_equal(numpy(t), a2)

    # I32

    def test_copy_np_to_ggml_1d_i32(self, ctx):
        t = lib.ggml_new_tensor_1d(ctx, lib.GGML_TYPE_I32, 10)
        a = np.arange(10, dtype=np.int32)
        self._test_copy_np_to_ggml(a, t)

    def test_copy_np_to_ggml_2d_i32(self, ctx):
        t = lib.ggml_new_tensor_2d(ctx, lib.GGML_TYPE_I32, 2, 3)
        a = np.arange(2 * 3, dtype=np.int32).reshape((2, 3))
        self._test_copy_np_to_ggml(a, t)

    def test_copy_np_to_ggml_3d_i32(self, ctx):
        t = lib.ggml_new_tensor_3d(ctx, lib.GGML_TYPE_I32, 2, 3, 4)
        a = np.arange(2 * 3 * 4, dtype=np.int32).reshape((2, 3, 4))
        self._test_copy_np_to_ggml(a, t)

    def test_copy_np_to_ggml_4d_i32(self, ctx):
        t = lib.ggml_new_tensor_4d(ctx, lib.GGML_TYPE_I32, 2, 3, 4, 5)
        a = np.arange(2 * 3 * 4 * 5, dtype=np.int32).reshape((2, 3, 4, 5))
        self._test_copy_np_to_ggml(a, t)

    def test_copy_np_to_ggml_4d_n_i32(self, ctx):
        dims = [2, 3, 4, 5] # GGML_MAX_DIMS is 4, going beyond would crash
        pdims = ffi.new('int64_t[]', len(dims))
        for i, d in enumerate(dims): pdims[i] = d
        t = lib.ggml_new_tensor(ctx, lib.GGML_TYPE_I32, len(dims), pdims)
        a = np.arange(np.prod(dims), dtype=np.int32).reshape(tuple(pdims))
        self._test_copy_np_to_ggml(a, t)

    # F32

    def test_copy_np_to_ggml_1d_f32(self, ctx):
        t = lib.ggml_new_tensor_1d(ctx, lib.GGML_TYPE_F32, 10)
        a = np.arange(10, dtype=np.float32)
        self._test_copy_np_to_ggml(a, t)

    def test_copy_np_to_ggml_2d_f32(self, ctx):
        t = lib.ggml_new_tensor_2d(ctx, lib.GGML_TYPE_F32, 2, 3)
        a = np.arange(2 * 3, dtype=np.float32).reshape((2, 3))
        self._test_copy_np_to_ggml(a, t)

    def test_copy_np_to_ggml_3d_f32(self, ctx):
        t = lib.ggml_new_tensor_3d(ctx, lib.GGML_TYPE_F32, 2, 3, 4)
        a = np.arange(2 * 3 * 4, dtype=np.float32).reshape((2, 3, 4))
        self._test_copy_np_to_ggml(a, t)

    def test_copy_np_to_ggml_4d_f32(self, ctx):
        t = lib.ggml_new_tensor_4d(ctx, lib.GGML_TYPE_F32, 2, 3, 4, 5)
        a = np.arange(2 * 3 * 4 * 5, dtype=np.float32).reshape((2, 3, 4, 5))
        self._test_copy_np_to_ggml(a, t)

    def test_copy_np_to_ggml_4d_n_f32(self, ctx):
        dims = [2, 3, 4, 5] # GGML_MAX_DIMS is 4, going beyond would crash
        pdims = ffi.new('int64_t[]', len(dims))
        for i, d in enumerate(dims): pdims[i] = d
        t = lib.ggml_new_tensor(ctx, lib.GGML_TYPE_F32, len(dims), pdims)
        a = np.arange(np.prod(dims), dtype=np.float32).reshape(tuple(pdims))
        self._test_copy_np_to_ggml(a, t)

    # F16

    def test_copy_np_to_ggml_1d_f16(self, ctx):
        t = lib.ggml_new_tensor_1d(ctx, lib.GGML_TYPE_F16, 10)
        a = np.arange(10, dtype=np.float16)
        self._test_copy_np_to_ggml(a, t)

    def test_copy_np_to_ggml_2d_f16(self, ctx):
        t = lib.ggml_new_tensor_2d(ctx, lib.GGML_TYPE_F16, 2, 3)
        a = np.arange(2 * 3, dtype=np.float16).reshape((2, 3))
        self._test_copy_np_to_ggml(a, t)

    def test_copy_np_to_ggml_3d_f16(self, ctx):
        t = lib.ggml_new_tensor_3d(ctx, lib.GGML_TYPE_F16, 2, 3, 4)
        a = np.arange(2 * 3 * 4, dtype=np.float16).reshape((2, 3, 4))
        self._test_copy_np_to_ggml(a, t)

    def test_copy_np_to_ggml_4d_f16(self, ctx):
        t = lib.ggml_new_tensor_4d(ctx, lib.GGML_TYPE_F16, 2, 3, 4, 5)
        a = np.arange(2 * 3 * 4 * 5, dtype=np.float16).reshape((2, 3, 4, 5))
        self._test_copy_np_to_ggml(a, t)

    def test_copy_np_to_ggml_4d_n_f16(self, ctx):
        dims = [2, 3, 4, 5] # GGML_MAX_DIMS is 4, going beyond would crash
        pdims = ffi.new('int64_t[]', len(dims))
        for i, d in enumerate(dims): pdims[i] = d
        t = lib.ggml_new_tensor(ctx, lib.GGML_TYPE_F16, len(dims), pdims)
        a = np.arange(np.prod(dims), dtype=np.float16).reshape(tuple(pdims))
        self._test_copy_np_to_ggml(a, t)

    # Mismatching shapes

    def test_copy_mismatching_shapes_1d(self, ctx):
        t = lib.ggml_new_tensor_1d(ctx, lib.GGML_TYPE_F32, 10)
        a = np.arange(10, dtype=np.float32)
        copy(a, t) # OK
        
        a = a.reshape((5, 2))
        with raises(AssertionError): copy(a, t)
        with raises(AssertionError): copy(t, a)
            
    def test_copy_mismatching_shapes_2d(self, ctx):
        t = lib.ggml_new_tensor_2d(ctx, lib.GGML_TYPE_F32, 2, 3)
        a = np.arange(6, dtype=np.float32)
        copy(a.reshape((2, 3)), t) # OK
        
        a = a.reshape((3, 2))
        with raises(AssertionError): copy(a, t)
        with raises(AssertionError): copy(t, a)

    def test_copy_mismatching_shapes_3d(self, ctx):
        t = lib.ggml_new_tensor_3d(ctx, lib.GGML_TYPE_F32, 2, 3, 4)
        a = np.arange(24, dtype=np.float32)
        copy(a.reshape((2, 3, 4)), t) # OK
        
        a = a.reshape((2, 4, 3))
        with raises(AssertionError): copy(a, t)
        with raises(AssertionError): copy(t, a)

    def test_copy_mismatching_shapes_4d(self, ctx):
        t = lib.ggml_new_tensor_4d(ctx, lib.GGML_TYPE_F32, 2, 3, 4, 5)
        a = np.arange(24*5, dtype=np.float32)
        copy(a.reshape((2, 3, 4, 5)), t) # OK
        
        a = a.reshape((2, 3, 5, 4))
        with raises(AssertionError): copy(a, t)
        with raises(AssertionError): copy(t, a)

    def test_copy_f16_to_f32(self, ctx):
        t = lib.ggml_new_tensor_1d(ctx, lib.GGML_TYPE_F32, 1)
        a = np.array([123.45], dtype=np.float16)
        copy(a, t)
        np.testing.assert_allclose(lib.ggml_get_f32_1d(t, 0), 123.45, rtol=1e-3)

    def test_copy_f32_to_f16(self, ctx):
        t = lib.ggml_new_tensor_1d(ctx, lib.GGML_TYPE_F16, 1)
        a = np.array([123.45], dtype=np.float32)
        copy(a, t)
        np.testing.assert_allclose(lib.ggml_get_f32_1d(t, 0), 123.45, rtol=1e-3)

    def test_copy_f16_to_Q5_K(self, ctx):
        n = 256
        t = lib.ggml_new_tensor_1d(ctx, lib.GGML_TYPE_Q5_K, n)
        a = np.arange(n, dtype=np.float16)
        copy(a, t)
        np.testing.assert_allclose(a, numpy(t, allow_copy=True), rtol=0.05)

    def test_copy_Q5_K_to_f16(self, ctx):
        n = 256
        t = lib.ggml_new_tensor_1d(ctx, lib.GGML_TYPE_Q5_K, n)
        copy(np.arange(n, dtype=np.float32), t)
        a = np.arange(n, dtype=np.float16)
        copy(t, a)
        np.testing.assert_allclose(a, numpy(t, allow_copy=True), rtol=0.05)

    def test_copy_i16_f32_mismatching_types(self, ctx):
        t = lib.ggml_new_tensor_1d(ctx, lib.GGML_TYPE_F32, 1)
        a = np.arange(1, dtype=np.int16)
        with raises(NotImplementedError): copy(a, t)
        with raises(NotImplementedError): copy(t, a)

class TestTensorCopy:

    def test_copy_self(self, ctx):
        t = lib.ggml_new_i32(ctx, 42)
        copy(t, t)
        assert lib.ggml_get_i32_1d(t, 0) == 42

    def test_copy_1d(self, ctx):
        t1 = lib.ggml_new_tensor_1d(ctx, lib.GGML_TYPE_F32, 10)
        t2 = lib.ggml_new_tensor_1d(ctx, lib.GGML_TYPE_F32, 10)
        a = np.arange(10, dtype=np.float32)
        copy(a, t1)
        copy(t1, t2)
        assert np.allclose(a, numpy(t2))
        assert np.allclose(numpy(t1), numpy(t2))

class TestGraph:

    def test_add(self, ctx):
        n = 256
        ta = lib.ggml_new_tensor_1d(ctx, lib.GGML_TYPE_F32, n)
        tb = lib.ggml_new_tensor_1d(ctx, lib.GGML_TYPE_F32, n)
        tsum = lib.ggml_add(ctx, ta, tb)
        assert tsum.type == lib.GGML_TYPE_F32

        gf = ffi.new('struct ggml_cgraph*')
        lib.ggml_build_forward_expand(gf, tsum)

        a = np.arange(0, n, dtype=np.float32)
        b = np.arange(n, 0, -1, dtype=np.float32)
        copy(a, ta)
        copy(b, tb)

        lib.ggml_graph_compute_with_ctx(ctx, gf, 1)

        assert np.allclose(numpy(tsum, allow_copy=True), a + b)

class TestQuantization:

    def test_quantized_add(self, ctx):
        n = 256
        ta = lib.ggml_new_tensor_1d(ctx, lib.GGML_TYPE_Q5_K, n)
        tb = lib.ggml_new_tensor_1d(ctx, lib.GGML_TYPE_F32, n)
        tsum = lib.ggml_add(ctx, ta, tb)
        assert tsum.type == lib.GGML_TYPE_Q5_K

        gf = ffi.new('struct ggml_cgraph*')
        lib.ggml_build_forward_expand(gf, tsum)

        a = np.arange(0, n, dtype=np.float32)
        b = np.arange(n, 0, -1, dtype=np.float32)
        copy(a, ta)
        copy(b, tb)

        lib.ggml_graph_compute_with_ctx(ctx, gf, 1)

        unquantized_sum = a + b
        sum = numpy(tsum, allow_copy=True)

        diff = np.linalg.norm(unquantized_sum - sum, np.inf)
        assert diff > 4
        assert diff < 5



# Convert GPT-J-6B h5 transformer model to ggml format
#
# Load the model using GPTJForCausalLM.
# Iterate over all variables and write them to a binary file.
#
# For each variable, write the following:
#   - Number of dimensions (int)
#   - Name length (int)
#   - Dimensions (int[n_dims])
#   - Name (char[name_length])
#   - Data (float[n_dims])
#
# By default, the bigger matrices are converted to 16-bit floats.
# This can be disabled by adding the "use-f32" CLI argument.
#
# At the start of the ggml file we write the model parameters
# and vocabulary.
#

import sys
import struct
import json
import torch
import numpy as np

from transformers import GPTJForCausalLM

# ref: https://github.com/openai/gpt-2/blob/master/src/encoder.py
def bytes_to_unicode():
    """
    Returns list of utf-8 byte and a corresponding list of unicode strings.
    The reversible bpe codes work on unicode strings.
    This means you need a large # of unicode characters in your vocab if you want to avoid UNKs.
    When you're at something like a 10B token dataset you end up needing around 5K for decent coverage.
    This is a signficant percentage of your normal, say, 32K bpe vocab.
    To avoid that, we want lookup tables between utf-8 bytes and unicode strings.
    And avoids mapping to whitespace/control characters the bpe code barfs on.
    """
    bs = list(range(ord("!"), ord("~")+1))+list(range(ord("¡"), ord("¬")+1))+list(range(ord("®"), ord("ÿ")+1))
    cs = bs[:]
    n = 0
    for b in range(2**8):
        if b not in bs:
            bs.append(b)
            cs.append(2**8+n)
            n += 1
    cs = [chr(n) for n in cs]
    return dict(zip(bs, cs))

if len(sys.argv) < 3:
    print("Usage: convert-h5-to-ggml.py dir-model [use-f32]\n")
    print("  ftype == 0 -> float32")
    print("  ftype == 1 -> float16")
    sys.exit(1)

# output in the same directory as the model
dir_model = sys.argv[1]
fname_out = sys.argv[1] + "/ggml-model.bin"

with open(dir_model + "/vocab.json", "r", encoding="utf-8") as f:
    encoder = json.load(f)

with open(dir_model + "/added_tokens.json", "r", encoding="utf-8") as f:
    encoder_added = json.load(f)

with open(dir_model + "/config.json", "r", encoding="utf-8") as f:
    hparams = json.load(f)

# possible data types
#   ftype == 0 -> float32
#   ftype == 1 -> float16
#
# map from ftype to string
ftype_str = ["f32", "f16"]

ftype = 1
if len(sys.argv) > 2:
    ftype = int(sys.argv[2])
    if ftype < 0 or ftype > 1:
        print("Invalid ftype: " + str(ftype))
        sys.exit(1)
    fname_out = sys.argv[1] + "/ggml-model-" + ftype_str[ftype] + ".bin"


model = GPTJForCausalLM.from_pretrained(dir_model, low_cpu_mem_usage=True)
#print (model)

list_vars = model.state_dict()
#print (list_vars)

fout = open(fname_out, "wb")

fout.write(struct.pack("i", 0x67676d6c)) # magic: ggml in hex
fout.write(struct.pack("i", hparams["vocab_size"]))
fout.write(struct.pack("i", hparams["n_positions"]))
fout.write(struct.pack("i", hparams["n_embd"]))
fout.write(struct.pack("i", hparams["n_head"]))
fout.write(struct.pack("i", hparams["n_layer"]))
fout.write(struct.pack("i", hparams["rotary_dim"]))
fout.write(struct.pack("i", ftype))

byte_encoder = bytes_to_unicode()
byte_decoder = {v:k for k, v in byte_encoder.items()}

fout.write(struct.pack("i", len(encoder) + len(encoder_added)))

for key in encoder:
    text = bytearray([byte_decoder[c] for c in key])
    fout.write(struct.pack("i", len(text)))
    fout.write(text)

for key in encoder_added:
    text = bytearray([byte_decoder[c] for c in key])
    fout.write(struct.pack("i", len(text)))
    fout.write(text)

for name in list_vars.keys():
    data = list_vars[name].squeeze().numpy()
    print("Processing variable: " + name + " with shape: ", data.shape)

    # we don't need these
    if name.endswith("attn.masked_bias") or name.endswith(".attn.bias"):
        print("  Skipping variable: " + name)
        continue

    n_dims = len(data.shape);

    # ftype == 0 -> float32, ftype == 1 -> float16
    ftype_cur = 0;
    if ftype != 0:
        if name[-7:] == ".weight" and n_dims == 2:
            print("  Converting to float16")
            data = data.astype(np.float16)
            ftype_cur = 1
        else:
            print("  Converting to float32")
            data = data.astype(np.float32)
            ftype_cur = 0
    else:
        if data.dtype != np.float32:
            print("  Converting to float32")
            data = data.astype(np.float32)
            ftype_cur = 0

    # for efficiency - transpose these matrices:
    # (note - with latest ggml this is no longer more efficient, so disabling it)
    #  "transformer.h.*.mlp.fc_in.weight"
    #  "transformer.h.*.attn.out_proj.weight"
    #  "transformer.h.*.attn.q_proj.weight"
    #  "transformer.h.*.attn.k_proj.weight"
    #  "transformer.h.*.attn.v_proj.weight"
    #if name.endswith(".mlp.fc_in.weight")     or \
    #   name.endswith(".attn.out_proj.weight") or \
    #   name.endswith(".attn.q_proj.weight")   or \
    #   name.endswith(".attn.k_proj.weight")   or \
    #   name.endswith(".attn.v_proj.weight"):
    #    print("  Transposing")
    #    data = data.transpose()

    # header
    str = name.encode('utf-8')
    fout.write(struct.pack("iii", n_dims, len(str), ftype_cur))
    for i in range(n_dims):
        fout.write(struct.pack("i", data.shape[n_dims - 1 - i]))
    fout.write(str);

    # data
    data.tofile(fout)

fout.close()

print("Done. Output file: " + fname_out)
print("")



# Convert GPT-2 h5 transformer model to ggml format
#
# Load the model using GPT2Model.
# Iterate over all variables and write them to a binary file.
#
# For each variable, write the following:
#   - Number of dimensions (int)
#   - Name length (int)
#   - Dimensions (int[n_dims])
#   - Name (char[name_length])
#   - Data (float[n_dims])
#
# By default, the bigger matrices are converted to 16-bit floats.
# This can be disabled by adding the "use-f32" CLI argument.
#
# At the start of the ggml file we write the model parameters
# and vocabulary.
#

import sys
import struct
import json
import numpy as np
import re

from transformers import GPT2Model

# ref: https://github.com/openai/gpt-2/blob/master/src/encoder.py
def bytes_to_unicode():
    """
    Returns list of utf-8 byte and a corresponding list of unicode strings.
    The reversible bpe codes work on unicode strings.
    This means you need a large # of unicode characters in your vocab if you want to avoid UNKs.
    When you're at something like a 10B token dataset you end up needing around 5K for decent coverage.
    This is a signficant percentage of your normal, say, 32K bpe vocab.
    To avoid that, we want lookup tables between utf-8 bytes and unicode strings.
    And avoids mapping to whitespace/control characters the bpe code barfs on.
    """
    bs = list(range(ord("!"), ord("~")+1))+list(range(ord("¡"), ord("¬")+1))+list(range(ord("®"), ord("ÿ")+1))
    cs = bs[:]
    n = 0
    for b in range(2**8):
        if b not in bs:
            bs.append(b)
            cs.append(2**8+n)
            n += 1
    cs = [chr(n) for n in cs]
    return dict(zip(bs, cs))

if len(sys.argv) < 2:
    print("Usage: convert-h5-to-ggml.py dir-model [use-f32]\n")
    sys.exit(1)

# output in the same directory as the model
dir_model = sys.argv[1]
fname_out = sys.argv[1] + "/ggml-model.bin"

with open(dir_model + "/vocab.json", "r", encoding="utf-8") as f:
    encoder = json.load(f)

with open(dir_model + "/added_tokens.json", "r", encoding="utf-8") as f:
    encoder_added = json.load(f)

with open(dir_model + "/config.json", "r", encoding="utf-8") as f:
    hparams = json.load(f)

# use 16-bit or 32-bit floats
use_f16 = True
if len(sys.argv) > 2:
    use_f16 = False
    fname_out = sys.argv[1] + "/ggml-model-f32.bin"

model = GPT2Model.from_pretrained(dir_model, low_cpu_mem_usage=True)
#print (model)

list_vars = model.state_dict()
#print (list_vars)

fout = open(fname_out, "wb")

fout.write(struct.pack("i", 0x67676d6c)) # magic: ggml in hex
fout.write(struct.pack("i", hparams["vocab_size"]))
fout.write(struct.pack("i", hparams["n_positions"]))
fout.write(struct.pack("i", hparams["n_embd"]))
fout.write(struct.pack("i", hparams["n_head"]))
fout.write(struct.pack("i", hparams["n_layer"]))
#fout.write(struct.pack("i", hparams["rotary_dim"]))
fout.write(struct.pack("i", use_f16))

byte_encoder = bytes_to_unicode()
byte_decoder = {v:k for k, v in byte_encoder.items()}

fout.write(struct.pack("i", len(encoder) + len(encoder_added)))

for key in encoder:
    text = bytearray([byte_decoder[c] for c in key])
    fout.write(struct.pack("i", len(text)))
    fout.write(text)

for key in encoder_added:
    text = bytearray([byte_decoder[c] for c in key])
    fout.write(struct.pack("i", len(text)))
    fout.write(text)

for name in list_vars.keys():
    data = list_vars[name].squeeze().numpy()
    print("Processing variable: " + name + " with shape: ", data.shape)

    # we don't need these
    if name.endswith("attn.masked_bias") or name.endswith(".attn.bias"):
        print("  Skipping variable: " + name)
        continue

    n_dims = len(data.shape);

    # ftype == 0 -> float32, ftype == 1 -> float16
    ftype = 0;
    if use_f16:
        if name[-7:] == ".weight" and n_dims == 2:
            print("  Converting to float16")
            data = data.astype(np.float16)
            ftype = 1
        else:
            print("  Converting to float32")
            data = data.astype(np.float32)
            ftype = 0

    # for efficiency - transpose these matrices:
    #  "transformer.h.*.mlp.c_proj.weight
    if name.endswith(".mlp.c_proj.weight"):
        print("  Transposing")
        data = data.transpose()

    # rename headers to keep compatibility
    if name == "ln_f.weight":
        name = "model/ln_f/g"
    elif name == "ln_f.bias":
        name = "model/ln_f/b"
    elif name == "wte.weight":
        name = "model/wte"
    elif name == "wpe.weight":
        name = "model/wpe"
    elif re.match(r"h\.\d+\.ln_1\.weight", name):
        i = re.findall("\d+", name)[0]
        name = f"model/h{i}/ln_1/g"
    elif re.match(r"h\.\d+\.ln_1\.bias", name):
        i = re.findall("\d+", name)[0]
        name = f"model/h{i}/ln_1/b"
    elif re.match(r"h\.\d+\.attn\.c_attn\.weight", name):
        i = re.findall("\d+", name)[0]
        name = f"model/h{i}/attn/c_attn/w"
    elif re.match(r"h\.\d+\.attn\.c_attn\.bias", name):
        i = re.findall("\d+", name)[0]
        name = f"model/h{i}/attn/c_attn/b"
    elif re.match(r"h\.\d+\.attn\.c_proj\.weight", name):
        i = re.findall("\d+", name)[0]
        name = f"model/h{i}/attn/c_proj/w"
    elif re.match(r"h.\d+.attn.c_proj.bias", name):
        i = re.findall("\d+", name)[0]
        name = f"model/h{i}/attn/c_proj/b"
    elif re.match(r"h.\d+.ln_2.weight", name):
        i = re.findall("\d+", name)[0]
        name = f"model/h{i}/ln_2/g"
    elif re.match(r"h.\d+.ln_2.bias", name):
        i = re.findall("\d+", name)[0]
        name = f"model/h{i}/ln_2/b"
    elif re.match(r"h.\d+.mlp.c_fc.weight", name):
        i = re.findall("\d+", name)[0]
        name = f"model/h{i}/mlp/c_fc/w"
    elif re.match(r"h.\d+.mlp.c_fc.bias", name):
        i = re.findall("\d+", name)[0]
        name = f"model/h{i}/mlp/c_fc/b"
    elif re.match(r"h.\d+.mlp.c_proj.weight", name):
        i = re.findall("\d+", name)[0]
        name = f"model/h{i}/mlp/c_proj/w"
    elif re.match(r"h.\d+.mlp.c_proj.bias", name):
        i = re.findall("\d+", name)[0]
        name = f"model/h{i}/mlp/c_proj/b"
    else:
        print("Unrecognized variable name. %s", name)

    str = name.encode('utf-8')

    fout.write(struct.pack("iii", n_dims, len(str), ftype))
    for i in range(n_dims):
        fout.write(struct.pack("i", data.shape[n_dims - 1 - i]))
    fout.write(str);

    # data
    data.tofile(fout)

fout.close()

print("Done. Output file: " + fname_out)
print("")



import os
from transformers import AutoTokenizer

os.environ['TOKENIZERS_PARALLELISM'] = "false"

list_repo_hf  = ["databricks/dolly-v2-3b",           # dolly-v2 (3b, 7b, 12b models share the same tokenizer)
                 "gpt2",                             # gpt-2 (gpt2-xl, gpt2-large share the same tokenizer)
                 "uer/gpt2-chinese-cluecorpussmall", # gpt-2-chinese
                 "EleutherAI/gpt-j-6b",              # gpt-j
                 "EleutherAI/gpt-neox-20b",          # gpt-neox
                 "EleutherAI/polyglot-ko-1.3b",      # gpt-neox (polyglot-ko 5.8b and 12.8b share the same tokenizer")
                 "rinna/japanese-gpt-neox-3.6b",     # gpt-neox
                 # mpt-7b (uses gpt-neox-20b tokenizer)
                 "replit/replit-code-v1-3b",         # replit
                 "bigcode/starcoder",                # starcoder (huggingface-cli login required)
                 "openai/whisper-tiny"               # whisper (base, large, large-v2 share the same tokenizer)
                 ]

repo2ggml     = {"databricks/dolly-v2-3b"           : "dolly-v2",
                 "gpt2"                             : "gpt-2",
                 "uer/gpt2-chinese-cluecorpussmall" : "gpt-2-chinese",
                 "EleutherAI/gpt-j-6b"              : "gpt-j",
                 "EleutherAI/gpt-neox-20b"          : "gpt-neox",
                 "EleutherAI/polyglot-ko-1.3b"      : "polyglot-ko",
                 "rinna/japanese-gpt-neox-3.6b"     : "gpt-neox-japanese",
                 "replit/replit-code-v1-3b"         : "replit",
                 "bigcode/starcoder"                : "starcoder",
                 "openai/whisper-tiny"              : "whisper"}

repo2language = {"databricks/dolly-v2-3b"           : "english",
                 "gpt2"                             : "english",
                 "uer/gpt2-chinese-cluecorpussmall" : "chinese",
                 "EleutherAI/gpt-j-6b"              : "english",
                 "EleutherAI/gpt-neox-20b"          : "english",
                 "EleutherAI/polyglot-ko-1.3b"      : "korean",
                 "rinna/japanese-gpt-neox-3.6b"     : "japanese",
                 "replit/replit-code-v1-3b"         : "english",
                 "bigcode/starcoder"                : "english",
                 "openai/whisper-tiny"              : "english"}

delimeter = ": "
test_sentences = []
with open("test-cases.txt", "r") as f:
    lines = [l.rstrip() for l in f.readlines()]
    for l in lines:
        if delimeter in l:
            language = l[:l.index(delimeter)]
            sentence = l[l.index(delimeter) + len(delimeter):]
            test_sentences.append((language.lower(), sentence))

for repo in list_repo_hf:

    target_language = repo2language[repo]

    tokenizer = AutoTokenizer.from_pretrained(repo, trust_remote_code=True)

    tokens_hf = []
    for language, sentence in test_sentences:
        if language == target_language:
            tokens = tokenizer.convert_tokens_to_ids(tokenizer.tokenize(sentence))
            tokens_hf.append((sentence, tokens))

    save_txt = repo2ggml[repo] + ".txt"
    with open(save_txt, "w") as f:
        f.writelines([sentence + " => " + ",".join(str(t) for t in tokens) + "\n" for sentence, tokens in tokens_hf])



from ggml import lib, ffi
from ggml.utils import init, copy, numpy
import numpy as np

ctx = init(mem_size=12*1024*1024) # automatically freed when pointer is GC'd
n = 256
n_threads = 4

a = lib.ggml_new_tensor_1d(ctx, lib.GGML_TYPE_Q5_K, n)
b = lib.ggml_new_tensor_1d(ctx, lib.GGML_TYPE_F32, n) # can't both be quantized
sum = lib.ggml_add(ctx, a, b) # all zeroes for now. Will be quantized too!

# See cffi's doc on how to allocate native memory: it's very simple!
# https://cffi.readthedocs.io/en/latest/ref.html#ffi-interface
gf = ffi.new('struct ggml_cgraph*')
lib.ggml_build_forward_expand(gf, sum)

copy(np.array([i for i in range(n)], np.float32), a)
copy(np.array([i*100 for i in range(n)], np.float32), b)

lib.ggml_graph_compute_with_ctx(ctx, gf, n_threads)

print(numpy(a, allow_copy=True))
print(numpy(b))
print(numpy(sum, allow_copy=True))


"""
  Python bindings for the ggml library.

  Usage example:

      from ggml import lib, ffi
      from ggml.utils import init, copy, numpy
      import numpy as np

      ctx = init(mem_size=10*1024*1024)
      n = 1024
      n_threads = 4

      a = lib.ggml_new_tensor_1d(ctx, lib.GGML_TYPE_Q5_K, n)
      b = lib.ggml_new_tensor_1d(ctx, lib.GGML_TYPE_F32, n)
      sum = lib.ggml_add(ctx, a, b)

      gf = ffi.new('struct ggml_cgraph*')
      lib.ggml_build_forward_expand(gf, sum)

      copy(np.array([i for i in range(n)], np.float32), a)
      copy(np.array([i*100 for i in range(n)], np.float32), b)
      lib.ggml_graph_compute_with_ctx(ctx, gf, n_threads)

      print(numpy(sum, allow_copy=True))

  See https://cffi.readthedocs.io/en/latest/cdef.html for more on cffi.
"""

try:
    from ggml.cffi import ffi as ffi
except ImportError as e:
    raise ImportError(f"Couldn't find ggml bindings ({e}). Run `python regenerate.py` or check your PYTHONPATH.")

import os, platform

__exact_library = os.environ.get("GGML_LIBRARY")
if __exact_library:
    __candidates = [__exact_library]
elif platform.system() == "Windows":
    __candidates = ["ggml_shared.dll", "llama.dll"]
else:
    __candidates = ["libggml_shared.so", "libllama.so"]
    if platform.system() == "Darwin":
        __candidates += ["libggml_shared.dylib", "libllama.dylib"]

for i, name in enumerate(__candidates):
    try:
        # This is where all the functions, enums and constants are defined
        lib = ffi.dlopen(name)
    except OSError:
        if i < len(__candidates) - 1:
            continue
        raise OSError(f"Couldn't find ggml's shared library (tried names: {__candidates}). Add its directory to DYLD_LIBRARY_PATH (on Mac) or LD_LIBRARY_PATH, or define GGML_LIBRARY.")

# This contains the cffi helpers such as new, cast, string, etc.
# https://cffi.readthedocs.io/en/latest/ref.html#ffi-interface
ffi = ffi
